'' =================================================================================================
''
''   File....... jm_pwm8.spin
''   Purpose.... 8-channel, fixed-frequency pwm
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2022 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 30 AUG 2022
''
''   {$P1}
''
'' =================================================================================================


con { fixed io pins }

  RX1      = 31  { I }                                          ' serial / programming
  TX1      = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }


con

  MAX_HZ   = 1000                                               ' maximum pwm frequency


con

  CHANNELS = 8                                                  ' groups of 4
  LAST_CH  = CHANNELS - 1


obj

  gc : "jm_gamma8"


var

  long  cog

  long  setup
  long  p_bright

  long  brightness[CHANNELS >> 2]                               ' brightness levels


pub null

'' This is not an application


pub start(count, base)

'' Start PWM8 pwm driver; uses one cog
'' -- count is number of pwm outputs (1 to CHANNELS)
'' -- base is the first (lsb) pin to use

  return startx(count, base, MAX_HZ)                            ' pwm @ 1kHz


pub startx(count, base, freq)

'' Start PWM8 pwm driver; uses one cog
'' -- count is number of pwm outputs (1 to CHANNELS)
'' -- base is the first (lsb) pin to use
'' -- freq is the modulation frequency

  stop                                                          ' stop if running

  setup.byte[0] := 1 #> count <#8
  setup.byte[1] := base
  setup.word[1] := (clkfreq / (1 #> freq <# MAX_HZ)) >> 8       ' ticks in 1/256 period for freq
  p_bright      := @brightness

  cog := cognew(@pwm_8x, @setup) + 1                            ' start the pwm cog

  return cog


pub stop

'' Stops PWM8 driver; frees a cog

  set_all(0)

  if (cog)
    cogstop(cog - 1)
    cog := 0


pub running

'' Reports true if PWM cog is loaded

  return (cog > 0)


pub set(ch, level)

'' Sets channel to specified level

  if ((ch => 0) and (ch =< LAST_CH))
    brightness.byte[ch] := 0 #> level <# 255

  return brightness.byte[ch]


pub set_all(level)

'' Set all channels to same level

  level := 0 #> level <# 255                                    ' limit to byte value
  bytefill(@brightness, level, CHANNELS)

  return level


pub high(ch)

'' Sets channel to on

  if ((ch => 0) and (ch =< LAST_CH))
    brightness.byte[ch] := 255


pub low(ch)

'' Sets channel to off

  if ((ch => 0) and (ch =< LAST_CH))
    brightness.byte[ch] := 0


pub toggle(ch)

'' Inverts channel level

  if ((ch => 0) and (ch =< LAST_CH))
    if (byte[@brightness][ch] & $80)
      byte[@brightness][ch]  := 0
    else
      brightness.byte[ch] := 255


pub inc(ch)

'' Increment channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    if (brightness.byte[ch] < 255)
      ++brightness.byte[ch]

  return brightness.byte[ch]


pub inc_all | ch

'' Increment all channels

  repeat ch from 0 to LAST_CH
    if (brightness.byte[ch] < 255)
      ++brightness.byte[ch]


pub dec(ch)

'' Decrement channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    if (brightness.byte[ch] > 0)
      --brightness.byte[ch]

  return brightness.byte[ch]


pub dec_all | ch

'' Decrement all channels

  repeat ch from 0 to LAST_CH
    if (brightness.byte[ch] > 0)
      --brightness.byte[ch]


pub digital(value, mask) | ch

'' Sets enabled pins (in mask) to corresponding bit in value

  repeat ch from 0 to LAST_CH                                   ' loop through outputs
    if (mask & (1 << ch))                                       ' if channel enabled
      if (value & (1 << ch))                                    ' if on
        brightness.byte[ch] := 255                              '  set to full
      else
        brightness.byte[ch] := 0                                '  set to off


pub dcd(bit)

'' Sets output to decoded power-of-2 value

  if (bit < 0)                                                  ' if negative
    set_all(0)                                                  '  all off
  elseif (bit =< LAST_CH)                                       ' if valid
    digital(1 << bit, %1111_1111)                               '  convert to bit output


pub read(ch)

'' Returns channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    return brightness.byte[ch]
  else
    return -1


pub address

'' Returns hub address of brightness array
'' -- allows direct external control of PWM outputs

  return @brightness


pub gamma(level)

'' Returns gamma-corrected level

  return gc.gamma(level)


dat { 8ch pwm driver }

                        org       0

pwm_8x                  mov       t1, par                       ' point to setup
                        rdlong    t2, t1                        ' read

                        mov       chcount, t2                   ' extract count
                        and       chcount, #$0F

                        mov       ch0pin, t2                    ' extract lsb pin
                        shr       ch0pin, #8
                        and       ch0pin, #$1F

                        mov       bittime, t2                   ' extract bit timing
                        shr       bittime, #16

                        add       t1, #4
                        rdlong    hubch0, t1                    ' read hub addr of levels array

                        mov       t1, #%1111_1111               ' create mask for pins
                        mov       t2, #8
                        sub       t2, chcount           wz
        if_nz           shr       t1, t2                        ' removed unused
                        shl       t1, ch0pin                    ' align lsbs
                        mov       dira, t1                      ' set outputs

                        mov       bridx, #1                     ' reset brightness index

                        mov       timer, bittime                ' start bit timer
                        add       timer, cnt                    ' sync with system timer


                        ' read channels

pwm_loop                mov       chpntr, hubch0
                        rdlong    ch03levels, chpntr
                        add       chpntr, #4
                        rdlong    ch47levels, chpntr


                        ' unrolled for speed

ch0                     mov       chmask, #%0000_0001           ' reset channel output mask
                        shl       chmask, ch0pin                ' align lsb
                        mov       t1, ch03levels                ' get byte from ch 0..3 group
                        and       t1, #$FF                      ' isolate
                        cmp       bridx, t1             wc      ' index past target level?
                        muxc      outa, chmask
ch0_exit                cmp       chcount, #1           wz      ' done?
        if_e            jmp       #loop_pad


ch1                     shl       chmask, #1
                        mov       t1, ch03levels
                        shr       t1, #8                        ' shift byte to lsb
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch1_exit                cmp       chcount, #2           wz
        if_e            jmp       #loop_pad


ch2                     shl       chmask, #1
                        mov       t1, ch03levels
                        shr       t1, #16
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch2_exit                cmp       chcount, #3           wz
        if_e            jmp       #loop_pad


ch3                     shl       chmask, #1
                        mov       t1, ch03levels
                        shr       t1, #24
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch3_exit                cmp       chcount, #4           wz
        if_e            jmp       #loop_pad


ch4                     shl       chmask, #1
                        mov       t1, ch47levels                ' get byte from ch 4..7 group
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch4_exit                cmp       chcount, #5           wz
        if_e            jmp       #loop_pad


ch5                     shl       chmask, #1
                        mov       t1, ch47levels
                        shr       t1, #8
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch5_exit                cmp       chcount, #6           wz
        if_e            jmp       #loop_pad


ch6                     shl       chmask, #1
                        mov       t1, ch47levels
                        shr       t1, #16
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask
ch6_exit                cmp       chcount, #7           wz
        if_e            jmp       #loop_pad


ch7                     shl       chmask, #1
                        mov       t1, ch47levels
                        shr       t1, #24
                        and       t1, #$FF
                        cmp       bridx, t1             wc
                        muxc      outa, chmask


loop_pad                waitcnt   timer, bittime

                        add       bridx, #1                     ' update brightness index
                        and       bridx, #$FF           wz      ' test rollover
        if_z            add       bridx, #1                     ' if cycle done, reset count
        if_z            jmp       #pwm_loop                     ' re-read pwm values

                        jmp       #ch0                          ' back to first output


' --------------------------------------------------------------------------------------------------

chcount                 res       1                             ' channel count
ch0pin                  res       1                             ' channel 0 pin #
bittime                 res       1                             ' timing for each bit
hubch0                  res       1                             ' hub address of brightness array

bridx                   res       1                             ' test brightness
timer                   res       1                             ' bit timer

chpntr                  res       1                             ' hub pointer to levels
chmask                  res       1                             ' channel mask

ch03levels              res       1                             ' levels for ch0..ch3
ch47levels              res       1                             ' levels for ch4..ch7

t1                      res       1                             ' work vars
t2                      res       1
t3                      res       1

                        fit       496


con { license }

{{

  MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}